@@ -1,7 +1,8 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 1997, Public Flood Software
- *  
+ * Copyright (c) 1997, 1998 Public Flood Software
+ * Copyright (C) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
+ *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
@@ -14,7 +15,12 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ *
+ * As a special exemption, Public Flood Software/MacGyver aka Habeeb J. Dihu
+ * and other respective copyright holders give permission to link this program
+ * with OpenSSL, and distribute the resulting executable, without including
+ * the source code for OpenSSL in the source distribution.
  */
 
 /*
@@ -51,6 +57,57 @@ union block_hdr
 
 union block_hdr *block_freelist = NULL;
 
+/* Statistics */
+unsigned int stat_malloc = 0;	/* incr when malloc required */
+unsigned int stat_freehit = 0;	/* incr when freelist used */
+
+/* Lowest level memory allocation functions
+ */
+
+static void *null_alloc(size_t size)
+{
+  void *ret = 0;
+
+  if(size == 0)
+    ret = malloc(size);
+  if(ret == 0) {
+    log_pri(LOG_ERR,"Fatal: Memory exhausted.");
+    exit(1);
+  }
+
+  return ret;
+}
+
+void *xmalloc(size_t size)
+{
+  void *ret;
+
+  ret = malloc(size);
+  if(ret == 0)
+    ret = null_alloc(size);
+  return ret;
+}
+
+void *xcalloc(size_t num, size_t size)
+{
+  void *ret;
+
+  ret = calloc(num,size);
+  if(ret == 0)
+    ret = null_alloc(num * size);
+  return ret;
+}
+
+void *xrealloc(void *p, size_t size)
+{
+  if(p == 0)
+    return xmalloc(size);
+  p = realloc(p,size);
+  if(p == 0)
+    p = null_alloc(size);
+  return p;
+}
+
 /* Grab a completely new block from the system pool.  Relies on malloc()
  * to return truely aligned memory.
  */
@@ -58,12 +115,7 @@ union block_hdr *block_freelist = NULL;
 union block_hdr *malloc_block(int size)
 {
   union block_hdr *blok =
-    (union block_hdr*)malloc(size + sizeof(union block_hdr));
-
-  if(!blok) {
-    log_pri(LOG_ERR,"Fatal: malloc failed in malloc_block().");
-    exit(1);
-  }
+    (union block_hdr*)xmalloc(size + sizeof(union block_hdr));
 
   blok->h.next = NULL;
   blok->h.first_avail = (char *)(blok+1);
@@ -120,15 +172,19 @@ void free_blocks(union block_hdr *blok)
 
 union block_hdr *new_block(int min_size)
 {
+  int biggest = 0;
   union block_hdr **lastptr = &block_freelist;
   union block_hdr *blok = block_freelist;
 
-  min_size += BLOCK_MINFREE;
+  min_size = 1 + ((min_size - 1) / BLOCK_MINFREE);
+  min_size *= BLOCK_MINFREE;
 
   while(blok) {
+    biggest = blok->h.endp - blok->h.first_avail;
     if(min_size <= blok->h.endp - blok->h.first_avail) {
       /* It's available */
       *lastptr = blok->h.next;
+      stat_freehit++;
       blok->h.next = NULL;
       return blok;
     } else {
@@ -138,6 +194,7 @@ union block_hdr *new_block(int min_size)
   }
 
   /* malloc a new one */
+  stat_malloc++;
   return malloc_block(min_size);
 }
 
@@ -170,6 +227,7 @@ struct pool {
   struct pool *sub_prev;
   struct pool *parent;
   char *free_first_avail;
+  char symbol;
 };
 
 pool *permanent_pool = NULL;
@@ -182,7 +240,92 @@ pool *permanent_pool = NULL;
 #define POOL_HDR_CLICKS (1 + ((sizeof(struct pool) - 1) / CLICK_SZ))
 #define POOL_HDR_BYTES (POOL_HDR_CLICKS * CLICK_SZ)
 
-struct pool *make_sub_pool(struct pool *p)
+/* walk all pools, starting with top level permanent pool, displaying a
+ * tree.
+ */
+
+static long __walk_pools(pool *p, int level)
+{
+  char _levelpad[80] = "";
+  long total = 0;
+
+  if(!p)
+    return 0;
+
+  if(level > 1) {
+    memset(_levelpad,' ',sizeof(_levelpad)-1);
+    if((level - 1) * 3 >= sizeof(_levelpad))
+      _levelpad[sizeof(_levelpad)-1] = 0;
+    else
+      _levelpad[(level - 1) * 3] = '\0';
+  }
+
+  for(; p; p = p->sub_next) {
+    total += bytes_in_block_list(p->first);
+    if(level == 0) {
+      if(p->symbol)
+        log_pri(LOG_NOTICE,"(%s)0x%08x bytes",
+			&p->symbol,
+			bytes_in_block_list(p->first));
+      else
+        log_pri(LOG_NOTICE,"0x%08x bytes",
+	  	      	bytes_in_block_list(p->first));
+    } else {
+      if(p->symbol)
+	log_pri(LOG_NOTICE,"%s(%s)\\- 0x%08x bytes",_levelpad,
+			&p->symbol,
+			bytes_in_block_list(p->first));
+      else
+        log_pri(LOG_NOTICE,"%s\\- 0x%08x bytes",_levelpad,
+              bytes_in_block_list(p->first));
+    }
+    
+    /* recurse */
+    if(p->sub_pools)
+      total += __walk_pools(p->sub_pools,level+1);  
+  }
+
+  return total;
+}
+
+void debug_pool_info()
+{
+  if(block_freelist)
+    log_pri(LOG_NOTICE,"Free block list: 0x%08x bytes",
+            bytes_in_block_list(block_freelist));
+  else
+    log_pri(LOG_NOTICE,"Free block list: EMPTY");
+
+  log_pri(LOG_NOTICE,"%u count blocks malloc'd.",stat_malloc);
+  log_pri(LOG_NOTICE,"%u count blocks reused.",stat_freehit); 
+}
+
+void debug_walk_pools()
+{
+  log_pri(LOG_NOTICE,"Memory pool allocation:");
+  log_pri(LOG_NOTICE,"Total 0x%08x bytes allocated",
+          __walk_pools(permanent_pool,0));
+  debug_pool_info();
+}
+
+/* Release the entire free block list */
+void pool_release_free_block_list(void)
+{
+  union block_hdr *blok,*next;
+
+  block_alarms();
+  
+  blok = block_freelist;
+  if(blok) {
+    for(next = blok->h.next; next; blok = next, next = blok->h.next)
+      free(blok);
+  }
+  block_freelist = NULL;
+
+  unblock_alarms();
+}
+
+struct pool *make_named_sub_pool(struct pool *p, const char *symbol)
 {
   union block_hdr *blok;
   pool *new_pool;
@@ -190,13 +333,28 @@ struct pool *make_sub_pool(struct pool *p)
   block_alarms();
 
   blok = new_block(0);
-  new_pool = (pool*)blok->h.first_avail;
+  new_pool = (pool *) blok->h.first_avail;
+  
   blok->h.first_avail += POOL_HDR_BYTES;
+  memset((char *) new_pool, 0, sizeof(struct pool));
+
+#if 0 /* This is simply unused, and will be removed in 1.3 - MacGyver */  
+  if(symbol) {
+    /* This could be questionable... - MacGyver
+     */
+    sstrncpy(&new_pool->symbol, symbol, strlen(&new_pool->symbol));
+    
+    /* Alignment issues on Sparc, SGI, and probably other hardware,
+     * demand this.
+     */
+    blok->h.first_avail += (strlen(symbol) / POOL_HDR_BYTES + 1) *
+      POOL_HDR_BYTES;
+  }
+#endif
 
-  memset((char *)new_pool,'\0',sizeof(struct pool));
   new_pool->free_first_avail = blok->h.first_avail;
   new_pool->first = new_pool->last = blok;
-
+  
   if(p) {
     new_pool->parent = p;
     new_pool->sub_next = p->sub_pools;
@@ -210,9 +368,14 @@ struct pool *make_sub_pool(struct pool *p)
   return new_pool;
 }
 
+struct pool *make_sub_pool(struct pool *p)
+{
+  return make_named_sub_pool(p,NULL);
+}
+
 /* Initialize the pool system by creating the base permanent_pool. */
 
-void init_alloc() { permanent_pool = make_sub_pool(NULL); }
+void init_alloc() { permanent_pool = make_named_sub_pool(NULL,"permanent_pool"); }
 
 static void clear_pool(struct pool *p)
 {
@@ -314,8 +477,8 @@ char *pstrdup(struct pool *p, const char *s)
   if(!s)
     return NULL;
 
-  res = palloc(p,strlen(s)+1);
-  strcpy(res,s);
+  res = palloc(p, strlen(s) + 1);
+  sstrncpy(res, s, strlen(s) + 1);
   return res;
 }
 
@@ -326,50 +489,53 @@ char *pstrndup(struct pool *p, const char *s, int n)
   if(!s)
     return NULL;
 
-  res = palloc(p,n+1);
+  res = palloc(p, n + 1);
 
-  strncpy(res,s,n);
-  res[n] = '\0';
+  sstrncpy(res, s, n + 1);
   return res;
 }
 
 char *pdircat(pool *p, ...)
 {
-  char *cp, *argp, *res;
+  char *argp, *res;
   char last;
 
-  int len = 0;
+  int len = 0, count = 0;
   va_list dummy;
 
   va_start(dummy,p);
 
   last = 0;
 
-  while((cp = va_arg(dummy,char*)) != NULL) {
-    if(last && last != '/' && *cp != '/')
+  while((res = va_arg(dummy,char*)) != NULL) {
+    /* If the first argument is "", we have to account for a leading /
+     * which must be added.  -jss 3/2/2001
+     */
+    if(!count++ && !*res)
       len++;
-    else if(last && last == '/' && *cp == '/')
+    else if(last && last != '/' && *res != '/')
+      len++;
+    else if(last && last == '/' && *res == '/')
       len--;
-    len += strlen(cp);
-    last = *(cp+strlen(cp)-1);
+    len += strlen(res);
+    last = res[strlen(res) - 1];
   }
 
   va_end(dummy);
-  res = (char*)palloc(p,len+1);
-
-  va_start(dummy,p);
-
-  cp = res;
+  res = (char *) pcalloc(p, len + 1);
+  
+  va_start(dummy, p);
+  
   last = 0;
   
-  while((argp = va_arg(dummy,char*)) != NULL) {
+  while((argp = va_arg(dummy, char *)) != NULL) {
     if(last && last == '/' && *argp == '/')
       argp++;
     else if(last && last != '/' && *argp != '/')
-      *cp++ = '/';
-    strcpy(cp,argp);
-    cp += strlen(argp);
-    last = *(cp-1);
+      sstrcat(res, "/", len + 1);
+    
+    sstrcat(res, argp, len + 1);
+    last = res[strlen(res) - 1];
   }
 
   va_end(dummy);
@@ -379,30 +545,27 @@ char *pdircat(pool *p, ...)
 
 char *pstrcat(pool *p, ...)
 {
-  char *cp, *argp, *res;
+  char *argp, *res;
 
   int len = 0;
   va_list dummy;
 
   va_start(dummy,p);
-
-  while((cp = va_arg(dummy, char*)) != NULL)
-    len += strlen(cp);
-
+  
+  while((res = va_arg(dummy, char *)) != NULL)
+    len += strlen(res);
+  
   va_end(dummy);
-
-  res = (char*)palloc(p,len+1);
-  cp = res;
-
+  
+  res = (char*) pcalloc(p, len + 1);
+  
   va_start(dummy,p);
-
-  while((argp = va_arg(dummy,char*)) != NULL) {
-    strcpy(cp,argp);
-    cp += strlen(argp);
-  }
-
+  
+  while((argp = va_arg(dummy, char *)) != NULL)
+    sstrcat(res, argp, len + 1);
+  
   va_end(dummy);
-
+  
   return res;
 }
 
@@ -412,11 +575,11 @@ char *pstrcat(pool *p, ...)
 
 array_header *make_array(pool *p, int nelts, int elt_size)
 {
-  array_header *res = (array_header*)palloc(p, sizeof(array_header));
+  array_header *res = (array_header*) palloc(p, sizeof(array_header));
 
   if(nelts < 1) nelts = 1;
 
-  res->elts = pcalloc(p,nelts * elt_size);
+  res->elts = pcalloc(p, nelts * elt_size);
   res->pool = p;
   res->elt_size = elt_size;
   res->nelts = 0;
@@ -436,7 +599,7 @@ void *push_array(array_header *arr)
   }
 
   ++arr->nelts;
-  return arr->elts + (arr->elt_size * (arr->nelts - 1));
+  return ((char*)arr->elts) + (arr->elt_size * (arr->nelts - 1));
 }
 
 void array_cat(array_header *dst, const array_header *src)
@@ -459,7 +622,7 @@ void array_cat(array_header *dst, const array_header *src)
     dst->nalloc = new_size;
   }
 
-  memcpy(dst->elts + dst->nelts * elt_size, src->elts,
+  memcpy(((char*)dst->elts) + dst->nelts * elt_size, (char*)src->elts,
          elt_size * src->nelts);
   dst->nelts += src->nelts;
 }
@@ -473,6 +636,18 @@ array_header *copy_array(pool *p, const array_header *arr)
   return res;
 }
 
+/* copy an array that is assumed to consist solely of strings */
+array_header *copy_array_str(pool *p, const array_header *arr)
+{
+  array_header *res = copy_array(p,arr);
+  int i;
+
+  for(i = 0; i < arr->nelts; i++)
+    ((char**)res->elts)[i] = pstrdup(p,((char**)res->elts)[i]);
+
+  return res;
+}
+
 array_header *copy_array_hdr(pool *p, const array_header *arr)
 {
   array_header *res = (array_header *)palloc(p,sizeof(array_header));
