@@ -1,25 +1,30 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 1997, Public Flood Software
- *  
+ * Copyright (c) 1997, 1998 Public Flood Software
+ *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * BUT witHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ *
+ * As a special exemption, Public Flood Software/MacGyver aka Habeeb J. Dihu
+ * and other respective copyright holders give permission to link this program
+ * with OpenSSL, and distribute the resulting executable, without including
+ * the source code for OpenSSL in the source distribution.
  */
 
 /* 
  * Timer system, based on alarm() and SIGALRM
- * $Id: timers.c,v 1.7 1997/07/08 20:00:30 flood Exp $
+ * $Id: timers.c,v 1.4 2001/06/18 17:12:45 flood Exp $
  */
 
 #include <signal.h>
@@ -27,12 +32,14 @@
 #include "conf.h"
 
 static int _current_timeout = 0;
+static int _total_time = 0;
 static int _sleep_sem = 0;
 static int alarms_blocked = 0,alarm_pending = 0;
 static xaset_t *timers = NULL;
 static xaset_t *recycled = NULL;
 static int _indispatch = 0;
 static int dynamic_timerno = 1024;
+static int _alarm_received = 0;
 
 xaset_t *free_timers = NULL;
 
@@ -111,8 +118,46 @@ static int _reset_timers(int elapsed)
 
   return (timers->xas_list ? ((timer_t*)timers->xas_list)->count : 0);
 }
-    
+
 void sig_alarm(int signum)
+{
+  struct sigaction act;
+
+  _alarm_received++;
+  act.sa_handler = sig_alarm;
+  sigemptyset(&act.sa_mask);
+  act.sa_flags = 0;
+#ifdef SA_INTERRUPT
+  act.sa_flags |= SA_INTERRUPT;
+#endif
+  sigaction(SIGALRM,&act,NULL);
+#ifdef HAVE_SIGINTERRUPT
+  siginterrupt(SIGALRM,1);
+#endif
+
+  /* Reset the alarm */
+  _total_time += _current_timeout;
+  if(_current_timeout)
+    alarm(_current_timeout);
+}
+
+void set_sig_alarm()
+{
+  struct sigaction act;
+
+  act.sa_handler = sig_alarm;
+  sigemptyset(&act.sa_mask);
+  act.sa_flags = 0;
+#ifdef SA_INTERRUPT
+  act.sa_flags |= SA_INTERRUPT;
+#endif
+  sigaction(SIGALRM,&act,NULL);
+#ifdef HAVE_SIGINTERRUPT
+  siginterrupt(SIGALRM,1);
+#endif
+}
+
+void handle_sig_alarm()
 {
   int new_timeout;
 
@@ -128,16 +173,18 @@ void sig_alarm(int signum)
    * called, if so, increment alarm_pending and exit swiftly
    */
 
-  if(!alarms_blocked) {
-    new_timeout = _current_timeout - alarm(0);
-    new_timeout = _reset_timers(new_timeout);
+  while(_alarm_received) {
+    _alarm_received = 0;
+    if(!alarms_blocked) {
+      new_timeout = _total_time + (_current_timeout - alarm(0));
+      _total_time = 0;
+      new_timeout = _reset_timers(new_timeout);
 
-    /*log_debug(DEBUG5,"alarm(%d)",new_timeout);*/
-    alarm(_current_timeout = new_timeout);
-  } else
-    alarm_pending++;
-
-  signal(SIGALRM,sig_alarm);
+      /*log_debug(DEBUG5,"alarm(%d)",new_timeout);*/
+      alarm(_current_timeout = new_timeout);
+    } else
+      alarm_pending++;
+  }
 }
 
 int reset_timer(int timerno, module *mod)
@@ -157,7 +204,8 @@ int reset_timer(int timerno, module *mod)
       t->count = t->interval;
       xaset_remove(timers,(xasetmember_t*)t);
       xaset_insert(recycled,(xasetmember_t*)t);
-      sig_alarm(SIGALRM);
+      _alarm_received++;
+      handle_sig_alarm();
       break;
     }
 
@@ -179,7 +227,8 @@ int remove_timer(int timerno, module *mod)
       else {
         xaset_remove(timers,(xasetmember_t*)t);
         xaset_insert(free_timers,(xasetmember_t*)t);
-        sig_alarm(SIGALRM);
+	_alarm_received++;
+        handle_sig_alarm();
       }      
       break;
     }
@@ -229,7 +278,9 @@ int add_timer(int seconds, int timerno, module *mod, callback_t cb)
     xaset_insert(recycled,(xasetmember_t*)t);
   } else {
     xaset_insert_sort(timers,(xasetmember_t*)t,TRUE);
-    sig_alarm(SIGALRM);
+    _alarm_received++;
+    set_sig_alarm();
+    handle_sig_alarm();
   }
 
   unblock_alarms();
@@ -253,7 +304,8 @@ void unblock_alarms()
   --alarms_blocked;
   if(alarms_blocked == 0 && alarm_pending) {
     alarm_pending = 0;
-    sig_alarm(SIGALRM);
+    _alarm_received++;
+    handle_sig_alarm();
   }
 }
 
@@ -278,8 +330,10 @@ int timer_sleep(int seconds)
     return -1;
 
   sigemptyset(&oset);
-  while(!_sleep_sem)
+  while(!_sleep_sem) {
     sigsuspend(&oset);
-
+    handle_sig_alarm();
+  }
+  
   return 0;  
 }
